
import shelve

from typing import List, Optional, Any

from ..json_io    import write_json
from ..dict_mixin import DictMixin

from .server import Server
from .client import Client


# ------------------------------------------------------------------------------
#
class Registry(Server):
    '''
    The `ru.zmq.Registry` is a ZMQ service which provides a hierarchical
    persistent data store.
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, url       : Optional[str] = None,
                       uid       : Optional[str] = None,
                       persistent: bool          = False) -> None:

        super().__init__(url=url, uid=uid)

        if persistent:
            self._data = shelve.open('%s.db' % self._uid, writeback=True)
        else:
            self._data = dict()

        self.register_request('put',  self.put)
        self.register_request('get',  self.get)
        self.register_request('keys', self.keys)
        self.register_request('del',  self.delitem)


    # --------------------------------------------------------------------------
    #
    def stop(self) -> None:

        if isinstance(self._data, dict):
            write_json(self._data, '%s.json' % self._uid)
        else:
            self._data.close()

        super().stop()


    # --------------------------------------------------------------------------
    #
    def put(self, key: str, val: Any) -> None:

        this  = self._data
        elems = key.split('.')

        for elem in elems[:-1]:
            step = this.get(elem)
            if step is None:
                step = dict()
                this[elem] = step
                this = this[elem]

        this[elems[-1]] = val

        if not isinstance(self._data, dict):
            self._data.sync()


    # --------------------------------------------------------------------------
    #
    def get(self, key: str) -> Optional[str]:

        try:
            this  = self._data
            elems = key.split('.')
            for elem in elems[:-1]:
                step = this.get(elem)
                if step is None:
                    return None
                this = step

            ret = this.get(elems[-1])
            return ret

        except AttributeError:
            return None


    # --------------------------------------------------------------------------
    #
    def keys(self) -> List[str]:

        return list(self._data.keys())


    # --------------------------------------------------------------------------
    #
    def delitem(self, key: str) -> None:

        del self._data[key]
        if not isinstance(self._data, dict):
            self._data.sync()


# ------------------------------------------------------------------------------
#
class RegistryClient(Client, DictMixin):
    '''
    The `ru.zmq.RegistryClient` class provides a simple dict-like interface to
    a remote `ru.zmq.Registry` service.  Note that only top-level dict-actions
    on the `RegistryClient` instance are synced with the remote service storage.
    '''


    # --------------------------------------------------------------------------
    #
    def __init__(self, url: str) -> None:

        super().__init__(url=url)


    # --------------------------------------------------------------------------
    # verbose API
    def get(self, key: str, default: Optional[str] = None) -> Optional[Any]:

        try:
            return self.request(cmd='get', key=key)
        except:
            return default


    def put(self, key: str,
                  val: Any) -> None:
        ret = self.request(cmd='put', key=key, val=val)
        assert ret is None
        return ret


    # --------------------------------------------------------------------------
    # dict mixin API
    def __getitem__(self, key: str) -> Optional[Any]:
        return self.get(key)


    def __setitem__(self, key: str, val: Any) -> None:
        return self.put(key, val)


    def __delitem__(self, key: str) -> None:
        ret = self.request(cmd='del', key=key)
        assert ret is None


    def keys(self) -> List[str]:
        ret = self.request(cmd='keys')
        assert isinstance(ret, list)
        return ret


# ------------------------------------------------------------------------------





import zmq
import msgpack

MODE_PUSH = 'push'
MODE_PULL = 'pull'


# ------------------------------------------------------------------------------
#
class Pipe(object):
    '''
    The `Pipe` class provides simple and direct n-to-m connectivity without
    load balancing or caching.  The class is expected to be used in the
    following way:

      - *one* endpoint (either push or pull) establishes the pipe:
        p = Pipe()
        p.connect_push()
        print(p.url)

      - all other endpoints (either push or pull) connect to the same pipe by
        specifying the URL from the first EP:
        p1 = Pipe()
        p1.connect_pull(url)

      - data should only be sent once at least one receiving EP is established
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, mode, url=None) -> None:
        '''
        Create a `Pipe` instance which can be used for either sending (`put()`)
        or receiving (`get()` / `get_nowait()`) data. according to the specified
        mode (`MODE_PUSH` or `MODE_PULL`).

        An URL can be specified for one end of the pipe - that end will then be
        in listening mode.  The other end of the pipe MUST use the connection
        URL provided by the listening end (`Pipe.url`).
        '''

        self._context = zmq.Context.instance()
        self._mode    = mode
        self._url     = None
        self._sock    = None
        self._poller  = zmq.Poller()

        if mode == MODE_PUSH:
            self._connect_push(url)

        elif mode == MODE_PULL:
            self._connect_pull(url)

        else:
            raise ValueError('unsupported pipe mode [%s]' % mode)


    # --------------------------------------------------------------------------
    #
    @property
    def url(self):
        return self._url


    # --------------------------------------------------------------------------
    #
    def _connect_push(self, url):
        '''
        Establish this pipe instance as sending endpoint.
        '''

        if self._sock:
            raise RuntimeError('already connected at %s' % self._url)

        if url:
            bind = False
        else:
            bind = True
            url  = 'tcp://*:*'

        self._sock = self._context.socket(zmq.PUSH)

        if bind: self._sock.bind(url)
        else   : self._sock.connect(url)

        self._url = self._sock.getsockopt(zmq.LAST_ENDPOINT)


    # --------------------------------------------------------------------------
    #
    def _connect_pull(self, url):
        '''
        Establish this Pipe as receiving endpoint.
        '''

        if self._sock:
            raise RuntimeError('already connected at %s' % self._url)

        if url:
            bind = False
        else:
            bind = True
            url  = 'tcp://*:*'

        self._sock = self._context.socket(zmq.PULL)

        if bind: self._sock.bind(url)
        else   : self._sock.connect(url)

        self._url = self._sock.getsockopt(zmq.LAST_ENDPOINT)
        self._poller.register(self._sock, zmq.POLLIN)


    # --------------------------------------------------------------------------
    #
    def put(self, msg):
        '''
        Send a message - if receiving endpoints are connected, exactly one of
        them will be able to receive that message.
        '''

        assert self._mode == MODE_PUSH
        self._sock.send(msgpack.packb(msg))


    # --------------------------------------------------------------------------
    #
    def get(self):
        '''
        Receive a message.  This call blocks until a message is available.
        '''

        assert self._mode == MODE_PULL
        return msgpack.unpackb(self._sock.recv())


    # --------------------------------------------------------------------------
    #
    def get_nowait(self, timeout: float = 0):
        '''
        Receive a message.  This call blocks for `timeout` seconds
        until a message is available.  If no message is available after timeout,
        `None` is returned.
        '''

        assert self._mode == MODE_PULL

        # zmq timeouts are in milliseconds
        socks = dict(self._poller.poll(timeout=int(timeout * 1000)))

        if self._sock in socks:
            return msgpack.unpackb(self._sock.recv())


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

__author__    = 'RADICAL-Cybertools Team'
__copyright__ = 'Copyright 2022, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import os

import radical.utils as ru

from unittest import TestCase


# ------------------------------------------------------------------------------
#
class TestDict(ru.TypedDict):
    pass


# ------------------------------------------------------------------------------
#
class ModulesTC(TestCase):

    # --------------------------------------------------------------------------
    #
    def test_get_type(self):

        for tname, t in [('str', str), ('int', int), ('float', float),
                         ('list', list), ('tuple', tuple), ('dict', dict)]:
            self.assertIs(ru.get_type(tname), t)

        class LocalTestClass:
            pass

        self.assertIs(ru.get_type('LocalTestClass'), LocalTestClass)

    # --------------------------------------------------------------------------
    #
    def test_load_class(self):

        fpath = '/tmp/file_with_py_class.py'

        with ru.ru_open(fpath, mode='w') as fd:
            fd.write("""
class NewClass:

    def hello(self):
        return 'hello'

""")
        obj = ru.load_class(fpath=fpath, cname='NewClass')()
        self.assertEqual(obj.hello(), 'hello')

        os.unlink(fpath)

    # --------------------------------------------------------------------------
    #
    def test_load_class_with_type(self):

        f = ru.load_class(fpath=__file__, cname='TestDict', ctype=ru.TypedDict)
        self.assertIsInstance(f(), ru.TypedDict)

# ------------------------------------------------------------------------------




#!/usr/bin/env python3

__author__    = 'Radical.Utils Development Team'
__copyright__ = 'Copyright 2021, RADICAL@Rutgers'
__license__   = 'MIT'


import time
import radical.utils as ru


# ------------------------------------------------------------------------------
#
def test_zmq_pipe():

    pipe_1 = ru.zmq.Pipe(ru.zmq.MODE_PUSH)

    url = pipe_1.url

    pipe_2 = ru.zmq.Pipe(ru.zmq.MODE_PULL, url)
    pipe_3 = ru.zmq.Pipe(ru.zmq.MODE_PULL, url)

    # let ZMQ settle
    time.sleep(1)

    for i in range(1000):
        pipe_1.put('foo %d' % i)

    result_2 = list()
    result_3 = list()

    for i in range(400):
        result_2.append(pipe_2.get())
        result_3.append(pipe_3.get())

    for i in range(100):
        result_2.append(pipe_2.get_nowait(timeout=1.0))
        result_3.append(pipe_3.get_nowait(timeout=1.0))

    assert len(result_2) == 500
    assert len(result_3) == 500

    test_2 = result_2.append(pipe_2.get_nowait(timeout=1.0))
    test_3 = result_3.append(pipe_3.get_nowait(timeout=1.0))

    assert test_2 is None
    assert test_3 is None


# ------------------------------------------------------------------------------
# run tests if called directly
if __name__ == '__main__':

    test_zmq_pipe()


# ------------------------------------------------------------------------------

